//---------------------------------------------------------------------------
// Copyright 2000 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/ft/ft.c,v $
// $Id: ft.c,v 1.30 2005/11/03 22:13:26 ah Exp $ 
//
// Description: print information about specified wav file
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: ft.c,v 1.30 2005/11/03 22:13:26 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
USING_NAMESPACE(AH)

#include <stdlib.h>
#ifdef NO_UNISTD
#   include <getopt.h>
#else
#   include <unistd.h>
#endif
#include <string.h>
#include <errno.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <drfftw.h> 
#include <iostream>
#include <fstream>
#include <vector>
#include <string>

#include <AH/Config/Types.h>
#include <AH/Etech/Types.h>
#include <AH/Etech/Riaa.h>
#include <AH/Etech/Algo.h>
#include <AH/IO/Riff/WaveFileCached.h>
#include <AH/Error/Class.h>
#include <AH/Error/Sys.h>
USING_NAMESPACE(AH)

#include "RedSpec.h"
#include "wisdom.h"

// --- locaL definitions ---

// --- local constants ---

static const char* cmdLineOptions = "D:LT:a:c:df:hlnrs:vz";
static const char* version = "1.5";
static const uint32 MAX_SAMPLES = 64 * 1024;  // 64k

// --- local data ---

static int debon = 0;
static bool riaa = false;
static unsigned int osize = 0;
static bool logsweep = false;
static bool normalize = false;
static bool zerofilling = false;
static double f_norm = -1;
static double db_add = 0;
static unsigned int channel = 1;
static bool linear_reduction = false;
//static char mode = '?';

// --- local classes ---

// --- local functions ---


static void usage()
{
    std::cerr << "Usage: ft [-" << cmdLineOptions << 
	"] inputfile [reference] outputfile\n"
	 << "       make a Fourier transformation.\n"
	 << "  options:\n"
	 << "    h         - print this help message\n"
	 << "    c <num>   - channel number (for multi channel WAV files)\n"
	 << "                stereo: 1=left 2=right (default=1)\n"
	 << "    d         - increase debug output (may be given more than once)\n"
	 << "    l         - linearise against logarithmic sweep\n"
	 << "    a <dB>    - add constant <dB> to curve\n"
	 << "    n         - normalize highest point = 0dB\n"
	 << "    f <freq>  - normalize at <freq> = 0dB\n"
	 << "    r         - linearise against RIAA preemphasis\n"
	 << "    s <num>   - reduce output size to <num>\n"
	 << "    v         - print version\n"
	 << "    z         - zerofilling to next power of 2\n"
	 << "    L         - reduce output size in linear mode (if -s)\n"
	 << "  input file  - WAV file or 16 bit mono PCM 44.1 kHz\n"
	 << "  reference   - WAV file or 16 bit mono PCM 44.1 kHz\n"
	 << "  output file - text file for gnuplot" << std::endl;
} // end of usage()



static void print_version()
{
    std::cerr << "ft Vs " << version << " " 
	 << "(c) 2002 Andreas Hnnebeck\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to andreas@huennebeck-online.de\n"
	 << "http://www.huennebeck-online.de"
	 << std::endl;
} // end of print_version()



static unsigned int read_pcm_file(double* data, unsigned int num, FILE* fpi)
{
    if (debon > 1) std::cerr << "read_file: read " << num << " points" << std::endl;
    unsigned int i = 0;

    int ret = 0;
    do
    {
	const unsigned int snum = 65536;
	int16 sbuf[snum];

	// set size of data to read on this turn:
	unsigned int rnum = snum;
	if (rnum > num) rnum = num;

	// read input buffer:
	if (debon > 2) std::cerr << "read_file: read " << rnum << " points" << std::endl;
	ret = fread(sbuf, sizeof(int16), rnum, fpi);
	if (debon > 2) std::cerr << "read_file: read returns " << ret << std::endl;
	if (-1 == ret) throw ErrorSys();
	for (int k = 0; k < ret; ++k)
	    data[i++] = sbuf[k];
	num -= ret;
	if ((unsigned int)ret < rnum) break;
    } while (num > 0);

    if (debon > 1) std::cerr << "read_file: returns " << i << std::endl;
    return i; // number of read entries
}



static uint32 read_wav_file(double* data, uint32 num, IORiffWaveFileCached& wave)
{
    if (debon > 1) std::cerr << "read_file: read " << num << " points" << std::endl;

    uint32 i = 0;
    for (uint32 curSam = 0; curSam < num; ++curSam)
    {
	int32 value = 0;
	if (0 >= wave.readSample(&value, curSam, channel))
	    break;
	data[i++] = value;
    }

    if (debon > 1) std::cerr << "read_file: returns " << i << std::endl;
    return i; // number of read entries
}



// return size of FT depending on zerofilling
static uint32 getInputSize(uint32 num)
{
    uint32 size = num;
    if (zerofilling)
    {
	// increase size to next power of 2 to speed up fft:
	int bits = (int)ceil(log((double)num)/log((double)2));
	size = 1 << bits;
	if (debon) std::cerr << "zerofilling: +" << size-num << " = " << size
			<< " points" << std::endl;
    }
    return size;
}



// get and check size of output file
static void getOutputSize(uint32 size)
{
    if (0 == osize) osize = size;
    if (osize > size)
    {
	std::cerr << "Error: wrong output size " << osize 
	     << " [<= " << size << "] "<< std::endl;
	exit(1);
    }
}



bool isWaveFile(const std::string& filename)
{
    const char* fn = filename.c_str();
    int len = strlen(fn);
    if (0 == strcasecmp(".wav", fn + len - 4))
	return true;
    return false;
}



// get normalization factor 
static double getNormFact(double f_inc, double f_max, 
			uint32 snum, const double* spectrum,
			const double* freq)
{
    double norm_fact = 0.0;
    if (!normalize) return norm_fact;

    if (f_norm > 0)
    {
	if (f_norm < f_max)
	{
	    // normalize to 0dB at f_norm:
	    if (debon) std::cerr << "normalize spectrum at f = " << f_norm 
			    << " Hz" << std::endl;

	    // find matching index:
	    uint32 idx = 0;
	    bool found = false;
	    for (unsigned int i = 0; i < snum; ++i)
	    {
		if (freq[i] > f_norm)
		{
		    found = true;
		    idx = i;
		    break;
		}
	    }
	    if (found)
	    {
		if (idx > 0)
		{
		    // check previous index, may be better
		    if ((f_norm - freq[idx-1]) < (freq[idx] - f_norm))
			--idx;
		}

		double val = spectrum[idx];
		norm_fact = -val;
		if (debon > 1)
		    std::cerr << "value at f = " << f_norm << " Hz = " 
			 << val << " dB, norm_fact = " 
			 << norm_fact << " dB" << std::endl;
	    }
	    else
	    {
		std::cerr << "Warning: normalization at f = " << f_norm 
		     << " Hz is not possible: f_max = " << f_max << " Hz"
		     << std::endl;
	    }
	}
	else std::cerr << "Warning: normalization at f = " << f_norm 
		  << " Hz is not possible: f_max = " << f_max << " Hz"
		  << std::endl;
		     
    }
    else
    {
	// normalize at highest point
	if (debon) std::cerr << "normalize spectrum at highest point" << std::endl;
	double max = 0;
	bool max_set = false;
	double f_peak = 0;
	for (unsigned int i = 0; i < snum; ++i)
	{
	    double val = spectrum[i];
	    if (!max_set || (val > max))
	    {
		max_set = true;
		max = val;
		f_peak = freq[i];
	    }
	}
	norm_fact = -max;
	if (debon > 1) 
	    std::cerr << "normalize spectrum at f_peak = " << f_peak 
		 << " Hz : peak = " << max << " dB, norm_fact = "
		 << norm_fact << " dB" << std::endl;
    }
    return norm_fact;
}



static Cplx* process_in(
    const string& in,
    FtWisdom& wis,
    uint32& num,          // number of samples
    uint32& sr,       // sample rate in Hz
    double& maxamp,
    uint32& snum
)
{
    //------------------------------------
    // read the data from the input file:
    //------------------------------------

    num = 0;          // number of samples
    sr = 44100;       // sample rate in Hz
    uint32 sb = 16;          // number of bits in sample
    uint32 size = 0;         // size of FT
    uint32 gotten = 0;       // number of samples read from file
    uint32 i = 0;            // loop counter
    fftw_real* din = 0;      // input samples (double)
    maxamp = (1 << (sb - 1)) - 1;

    if (isWaveFile(in))
    {
	IORiffWaveFileCached wave(in.c_str(), MAX_SAMPLES, debon);
	if (debon > 1) wave.printInfo(std::cerr);
	uint32 chnum = wave.getChannelNum();
	if (channel > chnum) 
	    throw ErrorBasic("illegal argument following option -c");

	num = wave.getSampleNum();
	sr = wave.getSampleRate();
	sb = wave.getSampleSize();
	maxamp = (1 << (sb - 1)) - 1;
	if (debon > 1) std::cerr << "Input: maxamp=" << maxamp << std::endl;

	size = getInputSize(num);

	// create input buffer
	din = new fftw_real[size];
	if (!din) throw ErrorSysOutOfMem();

	// read input data into input buffer:
	if (debon) std::cerr << "Input: reading wav file " << in << std::endl;
	gotten = read_wav_file(din, num, wave);
	if (((int)num - (int)gotten) > 0)
	    std::cerr << "WARNING: " << (int)num - (int)gotten 
		 << " missing data points in " << in << std::endl;
    }
    else // binary PCM file
    {
	// get size of file:
	struct stat st;
	int ret = stat(in.c_str(), &st);
	if (0 != ret) throw ErrorSys("Cannot stat " + in);

	// determine number of data points:
	num = st.st_size / sizeof(short);
	if (debon > 2) std::cerr << in << ": size=" << st.st_size << " = " << num
			    << " points" << std::endl;
	if (2 > num) throw ErrorBasic("Size of " + in + " is too small");

	size = getInputSize(num);

	// open data file for read:
	FILE* fpi = fopen(in.c_str(), "rb");
	if (!fpi) throw ErrorSys("Cannot open " + in);

	// create input buffer
	din = new fftw_real[size];
	if (!din)
	{
	    fclose(fpi);
	    throw ErrorSysOutOfMem();
	}

	// read input data into input buffer:
	if (debon) std::cerr << "Input: reading PCM file " << in << std::endl;
	gotten = read_pcm_file(din, num, fpi);

	// close file, done:
	fclose(fpi);
	if (((int)num - (int)gotten) > 0)
	    std::cerr << "WARNING: " << (int)num - (int)gotten 
		 << " missing data points in " << in << std::endl;
    }

    // make zerofilling:
    if (gotten > size)
	throw ErrorBasic("Error: Input: negative zerofilling <internal>");
    if (gotten < size)
    {
	if (!zerofilling) 
	    std::cerr << "Warning: ";
	if (!zerofilling || debon) 
	    std::cerr << "zerofilling of " << (int)size - (int)gotten 
		 << " point(s)" << std::endl;
	for (i = gotten; i < size; ++i)
	    din[i] = 0;
    }

    // determine min, mean and max value:
    double mean = 0;
    double max = 0;
    double min = 0;
    for (i = 0; i < gotten; ++i)
    {
	mean += din[i];
	if (din[i] > max) max = din[i];
	if (din[i] < min) min = din[i];
    }
    mean /= gotten;
    if (debon)
	std::cerr << "Input: TD Limits: min = " << min << ", mean = " << mean 
	     << ", max = " << max << std::endl;

    // remove DC offset:
    for (i = 0; i < gotten; ++i)
    {
	din[i] -= mean;
    }

    num = size;

    //----------------------------------------
    // calculate the FFT from the input data:
    //----------------------------------------
    if (debon) std::cerr << "Input: executing FFT" << std::endl;
    fftw_real* dout = new fftw_real[size];
    if (!dout)
    {
	delete[] din;
	throw ErrorSysOutOfMem();
    }

    rfftw_plan p = wis.getPlan(num);
    rfftw_one(p, din, dout);
    delete[] din; din = 0;

    // create complex spectrum data array
    if (debon) std::cerr << "Input: creating complex spectrum" << std::endl;
    snum = (num+1)/2;
    Cplx* spectrum = new Cplx[snum+1];
    if (!spectrum)
    {
	delete[] dout;
	throw ErrorSysOutOfMem();
    }

    // fill complex spectrum data array
    spectrum[0] = Cplx(mean, 0);
//    spectrum[0] = Cplx(dout[0], 0);
    for (i = 1; i < snum; ++i)
    {
	spectrum[i] = Cplx(dout[i], dout[num-i]) / snum;
    }
    if (num % 2)
    {
	spectrum[i] = Cplx(dout[num/2], 0) / snum;
    }
    delete[] dout; dout = 0;

    if (debon)
    {
	// determine min and max value:
	double smax = 0;
	double smin = 0;
	for (i = 0; i < snum; ++i)
	{
	    double val = abs(spectrum[i]);
	    if (val > smax) smax = val;
	    if (val < smin) smin = val;
	}
	std::cerr << "Input: SD Limits: min = " << smin << " max = " << smax 
	     << std::endl;
    }

    if (debon > 1) 
	std::cerr << "Input: apply option factors to spectrum" << std::endl;
    const double f_max = sr/2;
    const double f_inc = f_max/(double)(snum-1);
    double f = 0;
    for (unsigned int i = 0; i < snum; ++i, f += f_inc)
    {
	double k = 1.0 / maxamp;
	if (logsweep) k *= sqrt(f);

	spectrum[i] *= k;
	if (riaa) spectrum[i] *= RiaaPreEmphasis(f);
    }

    return spectrum;
}


static Cplx* process_ref(
    const string& ref,
    FtWisdom& wis,
    uint32 size,      
    uint32 sr
)
{
    //------------------------------------
    // read the data from the input file:
    //------------------------------------

    uint32 num = 0;
    uint32 sb = 16;          // number of bits in sample
    uint32 gotten = 0;       // number of samples read from file
    uint32 i = 0;            // loop counter
    fftw_real* din = 0;      // input samples (double)
    double maxamp = (1 << (sb - 1)) - 1;

    if (isWaveFile(ref))
    {
	IORiffWaveFileCached wave(ref.c_str(), MAX_SAMPLES, debon);
	if (debon > 1) wave.printInfo(std::cerr);
	uint32 chnum = wave.getChannelNum();
	if (channel > chnum) 
	    throw ErrorBasic("illegal argument following option -c");

	num = wave.getSampleNum();
	if (num > size)
	    throw ErrorBasic("Reference file size is too large");
	if (sr != wave.getSampleRate())
	    throw ErrorBasic("differing samplerate in reference file " + ref);
	sb = wave.getSampleSize();
	maxamp = (1 << (sb - 1)) - 1;
	if (debon > 1) std::cerr << "maxamp=" << maxamp << std::endl;

	// create input buffer
	din = new fftw_real[size];
	if (!din) throw ErrorSysOutOfMem();

	// read input data into input buffer:
	if (debon) std::cerr << "Reference: reading wav file " << ref << std::endl;
	gotten = read_wav_file(din, num, wave);
	if (((int)num - (int)gotten) > 0)
	    std::cerr << "WARNING: " << (int)num - (int)gotten 
		 << " missing data points in " << ref << std::endl;
    }
    else // binary PCM file
    {
	// get size of file:
	struct stat st;
	int ret = stat(ref.c_str(), &st);
	if (0 != ret) throw ErrorSys("Cannot stat " + ref);

	// determine number of data points:
	num = st.st_size / sizeof(short);
	if (num > size)
	    throw ErrorBasic("Reference file size is too large");
	if (sr != 44100)
	    throw ErrorBasic("differing samplerate in reference file " + ref);
	if (2 > num) throw ErrorBasic("Size of " + ref + " is too small");
	if (num > size)
	    throw ErrorBasic("Reference file size is too large");

	// open data file for read:
	FILE* fpi = fopen(ref.c_str(), "rb");
	if (!fpi) throw ErrorSys("Cannot open " + ref);

	// create input buffer
	din = new fftw_real[size];
	if (!din)
	{
	    fclose(fpi);
	    throw ErrorSysOutOfMem();
	}

	// read input data into input buffer:
	if (debon) std::cerr << "Reference: reading PCM file " << ref << std::endl;
	gotten = read_pcm_file(din, num, fpi);

	// close file, done:
	fclose(fpi);
	if (((int)num - (int)gotten) > 0)
	    std::cerr << "WARNING: " << (int)num - (int)gotten 
		 << " missing data points in " << ref << std::endl;
    }

    // make zerofilling:
    if (gotten > size)
	throw ErrorBasic("Error: Reference: negative zerofilling <internal>");
    if (gotten < size)
    {
	if (debon) 
	    std::cerr << "Reference: zerofilling of " << (int)size - (int)gotten 
		 << " point(s)" << std::endl;
	for (i = gotten; i < size; ++i)
	    din[i] = 0;
    }

    // determine min, mean and max value:
    double mean = 0;
    double max = 0;
    double min = 0;
    for (i = 0; i < gotten; ++i)
    {
	mean += din[i];
	if (din[i] > max) max = din[i];
	if (din[i] < min) min = din[i];
    }
    mean /= gotten;
    if (debon)
	std::cerr << "Reference: TD Limits: min = " << min << ", mean = " << mean 
	     << ", max = " << max << std::endl;

    // remove DC offset:
    for (i = 0; i < gotten; ++i)
    {
	din[i] -= mean;
    }

    num = size;

    //----------------------------------------
    // calculate the FFT from the input data:
    //----------------------------------------
    if (debon) std::cerr << "Reference: executing FFT" << std::endl;
    fftw_real* dout = new fftw_real[size];
    if (!dout)
    {
	delete[] din;
	throw ErrorSysOutOfMem();
    }

    rfftw_plan p = wis.getPlan(num);
    rfftw_one(p, din, dout);
    delete[] din; din = 0;

    // create complex spectrum data array
    if (debon) std::cerr << "Reference: creating complex spectrum" << std::endl;
    uint32 snum = (num+1)/2;
    Cplx* spectrum = new Cplx[snum+1];
    if (!spectrum)
    {
	delete[] dout;
	throw ErrorSysOutOfMem();
    }

    // fill complex spectrum data array
    spectrum[0] = Cplx(mean, 0);
    for (i = 1; i < snum; ++i)
    {
	spectrum[i] = Cplx(dout[i], dout[num-i]) / snum;
    }
    if (num % 2)
    {
	spectrum[i] = Cplx(dout[num/2], 0) / snum;
    }
    delete[] dout; dout = 0;

    if (debon)
    {
	// determine min and max value:
	double smax = 0;
	double smin = 0;
	for (i = 0; i < snum; ++i)
	{
	    double val = abs(spectrum[i]);
	    if (val > smax) smax = val;
	    if (val < smin) smin = val;
	}
	std::cerr << "Reference: SD Limits: min = " << smin << " max = " << smax 
	     << std::endl;
    }

    if (debon > 1) 
	std::cerr << "Reference: apply option factors to spectrum" << std::endl;
    const double f_max = sr/2;
    const double f_inc = f_max/(double)(snum-1);
    double f = 0;
    for (unsigned int i = 0; i < snum; ++i, f += f_inc)
    {
	double k = 1.0 / maxamp;
	if (logsweep) k *= sqrt(f);

	spectrum[i] *= k;
    }

    return spectrum;
}



static inline double ld(double val)
{
    return log(val) / log((double)2);
}



static void process(const string& in, const string& out, const string& ref)
{
    FtWisdom wis(debon);

    // make ft of input file:
    uint32 num = 0;          // number of samples
    uint32 sr = 0;       // sample rate in Hz
    double maxamp = 0;
    uint32 snum = 0;
    Cplx* spectrum = process_in(in, wis, num, sr, maxamp, snum);

    // reduce spectrum to desired output size
    getOutputSize(snum);
    const double f_max = sr/2;
    const double f_inc = f_max/(double)(snum-1);
    ReducedSpectrum sred(snum, osize, f_max, debon);
    uint32 bad_cnt = sred.process(spectrum, linear_reduction);
    if (debon && bad_cnt)
	std::cerr << bad_cnt << " bad value(s) in input spectrum" << std::endl;
    if (ref != "")
    {
	// make ft of reference file:
	Cplx* reference = process_ref(ref, wis, num, sr);
	// reduce spectrum to desired output size
	ReducedSpectrum refred(snum, osize, f_max, debon);
	bad_cnt = refred.process(reference, linear_reduction);
	if (debon && bad_cnt)
	    std::cerr << bad_cnt << " bad value(s) in reference spectrum" << std::endl;

	// make the difference spectrum of input and reference:
	if (debon > 1) 
	    std::cerr << "subtract reference spectrum from input spectrum" << std::endl;
	for (unsigned int i = 0; i < osize; ++i)
	{
	    sred.addMeanAbsVal(i, -refred.getMeanAbsVal(i));
	    sred.addMaxAbsVal(i, -refred.getMaxAbsVal(i));
	    sred.addMinAbsVal(i, -refred.getMinAbsVal(i));
	}
	delete[] reference;
    }

    // get normalisation factor if desired:
    double norm_fact_mean = getNormFact(f_inc, f_max, osize, 
					sred.getMeanAbs(), sred.getFreq());
    double norm_fact_max = norm_fact_mean;
    double norm_fact_min = norm_fact_mean;
    if (normalize && (osize < snum))
    {
	norm_fact_max = getNormFact(f_inc, f_max, osize, 
				    sred.getMaxAbs(), sred.getFreq());
	norm_fact_min = getNormFact(f_inc, f_max, osize, 
				    sred.getMinAbs(), sred.getFreq());
    }

    if (0 != db_add)
    {
	for (unsigned int i = 0; i < osize; ++i)
	{
	    sred.addMeanAbsVal(i, db_add);
	    sred.addMaxAbsVal(i, db_add);
	    sred.addMinAbsVal(i, db_add);
	}
    }

    //-----------------------------------------
    // write the results into the output file:
    //-----------------------------------------
    if (debon) std::cerr << "writing file " << out << std::endl;
    // open output file
    FILE* fpo = fopen(out.c_str(), "w");
    if (!fpo) throw ErrorSys("Cannot open " + out);
    // print comment line:
    fprintf(fpo, "#          ---- mean ----- ---- max ---- ---- min ---\n");
    fprintf(fpo, "# freq/Hz | dB    degree  | db   degree | db   degree\n");

    // print entries:
    for (unsigned int i = 0; i < osize; ++i)
    {
	// calculate mean value in dB and degree:
	double db_mean = sred.getMeanAbsVal(i);
	if (normalize) db_mean += norm_fact_mean;
	double phi_mean = inDeg(sred.getMeanPhiVal(i));
	double db_max = db_mean;
	double phi_max = phi_mean;
	double db_min = db_mean;
	double phi_min = phi_mean;

	if (osize < snum)
	{
	    // calculate max value in dB and degree:
	    db_max = sred.getMaxAbsVal(i);
	    if (normalize) db_max += norm_fact_max;
	    phi_max = inDeg(sred.getMaxPhiVal(i));

	    // calculate min value in dB and degree:
	    db_min = sred.getMinAbsVal(i);
	    if (normalize) db_min += norm_fact_min;
	    phi_min = inDeg(sred.getMinPhiVal(i));
	}
	fprintf(fpo, "%f\t%f %f\t%f %f\t%f %f\n", 
		sred.getFreqVal(i), db_mean, phi_mean, db_max, phi_max,
		db_min, phi_min);

    }
    fclose(fpo);
    delete[] spectrum;
}


// --- global functions ---


int main(int argc, char * const argv[])
{
    // loop through command line options:
    int option;
    while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
    {
	switch (option) 
	{
//	case 'A': 
//	case 'B': 
//	case 'C': 
//	case 'E': mode = option; break;

	case 'L': linear_reduction = true; break;
	case 'd': ++debon; break;
	case 'c': channel = atoi(optarg); break;
	case 'h': usage(); exit(0);
	case 'l': logsweep = true; break;
	case 'n': normalize = true; break;
	case 'r': riaa = true; break;
	case 'a': db_add = atof(optarg); break;
	case 'f': normalize = true; f_norm = atof(optarg); break;
	case 's': osize = atoi(optarg); break;
	case 'v': print_version(); exit(0);
	case 'z': zerofilling = true; break;
	default : usage(); exit(1);
	}
    }
    if (optind >= argc) 
    {
	std::cerr << "Error: missing file" << std::endl;
	usage();
	exit(1);
    }
    if (osize && (osize < 2))
    {
	std::cerr << "Error: wrong output size " << osize 
	     << " [>= 2] "<< std::endl;
	usage();
	exit(1);
    }

    // loop through the remaining arguments: these are hopefully raw PCM files
    vector<string> args;
    for (int i=optind; i < argc; i++)
    {
	args.push_back(argv[i]);
    }

    string in_file;
    string out_file;
    string ref_file;
    switch (args.size())
    {
    case 3:  in_file = args[0];
	     ref_file = args[1];
	     out_file = args[2];
	     break;
    case 2:  in_file = args[0];
	     out_file = args[1];
	     break;
    case 0:  std::cerr << "Error: missing input file" << std::endl;
	     usage();
	     exit(1); 
             // fall through
    case 1:  std::cerr << "Error: missing output file" << std::endl;
	     usage();
	     exit(1);
    default: std::cerr << "Error: too many arguments" << std::endl;
	     usage();
	     exit(1);
    }

    // process the files
    static unsigned int errnum = 0;
    try
    {
	process(in_file, out_file, ref_file); 
    }
    catch (ErrorClass& err)
    {
	++errnum;
	std::cerr << "Error: " << err.getMsg() << std::endl;
    }
    catch (exception& err)
    {
	++errnum;
	std::cerr << "Error: " << err.what() << std::endl;
    }
    
    exit (errnum ? 1 : 0);
    return  (errnum ? 1 : 0); // to keep MSVC quiet
} // end of main()


// --- end of file ---
