//---------------------------------------------------------------------------
// Copyright 2000 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/wavtool/wavcut.c,v $
// $Id: wavcut.c,v 1.11 2002/05/18 13:43:31 ah Exp $ 
//
// Description: cut a pice out of a wav file into a new wav file
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: wavcut.c,v 1.11 2002/05/18 13:43:31 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
USING_NAMESPACE(AH)

#include <stdlib.h>
#ifdef NO_UNISTD
#   include <getopt.h>
#else
#   include <unistd.h>
#endif
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
USING_NAMESPACE(std)

#include <AH/Error/Class.h>
#include <AH/Error/Basic.h>
#include <AH/IO/Convert.h>
#include <AH/IO/Riff/WaveFile.h>
USING_NAMESPACE(AH)

#include "fileset.h"
#include "datacopy.h"

// --- local constants ---

static const char* cmdLineOptions = "f:F:hi:s:S:t:T:vV";
static const char* version = "1.1";
static const uint32 MAX_BUF_SIZE = 1 * 1024 * 1024;  // 1MB

// --- local variables ---

static int verbose = 0;

// --- local classes ---

class AudioFormat
{
 public:
    AudioFormat()
	: channelNum(0), sampleSize(0), sampleRate(0), sampleNum(0) {}
    void addSampleNum(uint32 sn)
    {
	uint32 maxSampleNum = (0xffffffff - 0x200) / channelNum;
	uint32 maxAddNum = maxSampleNum - (sampleNum * channelNum);
	if (sn > maxAddNum) throw ErrorBasic();
	sampleNum += sn;
    }

    uint16 channelNum;	// number of channels
    uint16 sampleSize;	// size of sample in bits
    uint32 sampleRate;	// sample rate in Hz
    uint32 sampleNum;	// number of samples per channel
};	


// --- local functions ---


static void usage(bool help = false)
{
    cerr << "Usage: wavcut [-" << cmdLineOptions << "] file\n"
	 << "       cut a piece from WAV file(s) into new WAV file.\n"
	 << "  options:\n"
	 << "   -h                  - print this help message\n"
	 << "   -i <input file>     - specify input file\n"
	 << "   -f <time>           - fade in with specified time in sec\n"
	 << "   -F <time>           - fade out with specified time in sec\n"
	 << "   -s <sample-range>   - cut specified range of samples\n"
	 << "      e.g '-s 100..200'  cut samples 100 to 200 inclusively\n"
	 << "      e.g '-s 100..'     cut from sample 100 to end of file\n"
	 << "      e.g '-s ..200'     cut from begin of file to sample 200\n"
	 << "   -S <time>           - add silence time in sec\n"
	 << "   -t <time-range>     - cut specified time range\n"
	 << "                         time is given as Min:Sec.MilliSec\n"
	 << "      e.g '-t 0:12.3-3:15.23'\n"
	 << "      e.g '-t 0:12.3-'   cut from 0:12.3 to end of file\n"
	 << "      e.g '-t ..3:15.23' cut from begin of file to 3:15.23\n"
	 << "   -T <tocfile>        - create toc file for cdrdao\n"
	 << "   -v                  - verbose mode (increase debug level)\n"
	 << "   -V                  - print version" << endl;
    if (!help) return;

    cerr << "\n"
	 << "Examples:\n"
	 << "  wavcut -i a.wav -i b.wav c.wav\n"
	 << "  - concatenate files a.wav and b.wav into file c.wav\n"
	 << endl;
} // end of usage()


static void print_version()
{
    cerr << "wavcut Vs " << version << " " 
	 << "(c) 2000 Andreas Hnnebeck\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to andreas@huennebeck-online.de\n"
	 << "http://www.huennebeck-online.de"
	 << endl;
} // end of print_version()


//----------------------------------------------------------------
// check input files for existence and for matching audio format
// On Error: throws an exception
//----------------------------------------------------------------
static void check_input_files(const FileSettings& files, AudioFormat& fmt)
{
    if (2 < verbose) cerr << "check input files" << endl;

    double accumulated_silence = 0;
    double one_sec_samples = 0;

    for (unsigned int i=0; i < files.getNum(); i++)
    {
	double cur_silence = files.getSilenceTime(i);
	if (cur_silence > 0)
	{
	    accumulated_silence += cur_silence;
	    continue;
	}

	const string& inputFileName = files.getFileName(i);
	// create a wav file object and let it print some info
	if (3 < verbose) cerr << "check input file \"" << inputFileName
			      << "\":" << endl;
	IORiffWaveFile wave(inputFileName.c_str());
	if (0 == i)
	{
	    //-------------------------------------
	    // first file determines audio format:
	    // - catch format specifications
	    //-------------------------------------
	    fmt.channelNum = wave.getChannelNum();
	    fmt.sampleSize = wave.getSampleSize();
	    fmt.sampleRate = wave.getSampleRate();
	    // add number of samples needed for this file
	    fmt.sampleNum  = wave.getSampleNum();
	    one_sec_samples = wave.getSampleNum(1.0);
	    if (4 < verbose) 
		cerr << "- channels = " << fmt.channelNum << " / "
		     << fmt.sampleSize << " bits / "
		     << fmt.sampleRate << " Hz / "
		     << fmt.sampleNum << " samples" << endl;
	}
	else
	{
	    //----------------------------------------------------
	    // all other files should have the same audio format:
	    // - check number of channels:
	    //----------------------------------------------------
	    if (wave.getChannelNum() != fmt.channelNum)
	    {
		char buf[256];
		(void)sprintf(buf,
			      "Error: different audio format:\n"
			      " - is %d channel(s) in file %s\n"
			      " - expected %d channel(s) as in file %s",
			      wave.getChannelNum(), inputFileName.c_str(), 
			      fmt.channelNum, files.getFileName(0).c_str());
		throw ErrorBasic(buf);
	    }
	    //-------------------------
	    // - check size of sample:
	    //-------------------------
	    if (wave.getSampleSize() != fmt.sampleSize)
	    {
		char buf[256];
		(void)sprintf(buf,
			      "Error: different audio format:\n"
			      " - is %d bits/sample in file %s\n"
			      " - expected %d bits/sample as in file %s",
			      wave.getSampleSize(), inputFileName.c_str(), 
			      fmt.sampleSize, files.getFileName(0).c_str());
		throw ErrorBasic(buf);
	    }
	    //----------------------
	    // - check sample rate:
	    //----------------------
	    if (wave.getSampleRate() != fmt.sampleRate)
	    {
		char buf[256];
		(void)sprintf(buf,
			      "Error: different audio format:\n"
			      " - is %d samples/sec in file %s\n"
			      " - expected %d samples/sec as in file %s",
			      wave.getSampleRate(), inputFileName.c_str(), 
			      fmt.sampleRate, files.getFileName(0).c_str());
		throw ErrorBasic(buf);
	    }
	    // add number of samples needed for this file
	    fmt.addSampleNum(wave.getSampleNum());
	}
    }
    // add number of samples needed for silence
    uint32 num = (uint32)((double)one_sec_samples * accumulated_silence);
    fmt.addSampleNum(num);
} // end of check_input_files()



//----------------------------------------------------------------
// copy specified part of one input files into output file
// (based on time or samples)
// On Error: throws an exception
//----------------------------------------------------------------
static void read_one_file
(
    const FileSettings& files,
    int fileNum,
    IORiffWaveFile& out,
    const string& outputFileName,
    vector<uint32>& partSamples
)
{
    // create a wav file object:
    const string& inputFileName = files.getFileName(fileNum);
    if (3 < verbose) cerr << "open input file \"" << inputFileName
			  << "\":" << endl;
    IORiffWaveFile in(inputFileName.c_str());
    if (verbose  > 3) in.printInfo(cerr);

    // get desired range (samples or times):
    int s1;  // first sample
    int s2;  // sample after fade in is finished
    int s3;  // sample where fade out starts
    int s4;  // last sample
    files.getSampleRange(fileNum, s1, s4);
    if (0 > s1) s1 = 0;
    if (0 > s4) s4 = in.getSampleNum() - 1;

    double t1, t4;
    files.getTimeRange(fileNum, t1, t4);
    if ((0 <= t1) || (0 <= t4))
    {
	// time setting overrides sample setting:
	if (0 <= t1) s1 = in.getSampleNum(t1);
	else         s1 = 0;
	if (0 <= t4) s4 = in.getSampleNum(t4);
	else         s4 = in.getSampleNum() - 1;
    }

    s2 = s1;
    double t2 = files.getFadeInTime(fileNum);
    if (0 < t2) s2 += in.getSampleNum(t2);

    s3 = s4;
    double t3 = files.getFadeOutTime(fileNum);
    if (0 < t3) s3 -= in.getSampleNum(t3);

    if (s2 > s1)
    {
	// copy with fade in
	if (2 < verbose) cerr << "fade in from " << s1 << "-" << s2-1 << endl;
	DataCopyFadeIn proc(in, inputFileName, out, outputFileName, verbose);
	proc.copy(s1, s2-1);
    }

    // copy without manipulation
    if (2 < verbose) cerr << "copy from " << s2 << "-" << s3 << endl;
    DataCopyRaw proc(in, inputFileName, out, outputFileName, verbose);
    proc.copy(s2, s3);

    if (s4 > s3+1)
    {
	// copy with fade out
	if (2 < verbose) cerr << "fade out from " << s3+1 << "-" << s4 << endl;
	DataCopyFadeOut proc(in, inputFileName, out, outputFileName, verbose);
	proc.copy(s3+1, s4);
    }
    partSamples.push_back(s4 - s1 + 1);
} // end of read_one_file()



//----------------------------------------------------------------
// write specified silence time into output file
// On Error: throws an exception
//----------------------------------------------------------------
static void write_silence
(
    double silence,
    IORiffWaveFile& out,
    const string& outputFileName
)
{
    // get desired size of null buffer:
    uint32 sizeToWrite = out.getDataBufSize(silence);

    if (2 < verbose) cerr << "add " << sizeToWrite << " null bytes for " 
			  << silence << " sec. silence" << endl;

    // make a buffer for null data (on stack):
    const uint32 bufSize = MAX_BUF_SIZE;
    unsigned char buf[bufSize];
    memset(buf, 0, (sizeToWrite < bufSize ? sizeToWrite : bufSize));

    uint32 sizeWritten = 0;
    while (0 < sizeToWrite)
    {
	uint32 curSizeToWrite = (sizeToWrite < bufSize ? 
				 sizeToWrite : bufSize);
	if (4 < verbose) cerr << "write " << curSizeToWrite 
			      << " bytes into output file \"" 
			      << outputFileName << "\":" << endl;
	sizeWritten = out.writeRaw(buf, curSizeToWrite);
	if (0 >= sizeWritten) break;
	sizeToWrite -= sizeWritten;
    }
    if (sizeToWrite)
	cerr << "Warning: could not write " << sizeToWrite  
	     << " bytes into output file \"" 
	     << outputFileName << "\"" << endl;
} // end of write_silence()


string samples_to_toc(uint32 samples)
{
    uint32 minutes = samples / (44100 * 4 * 60);
    samples %= (44100 * 4 * 60);
    uint32 seconds = samples / (44100 * 4);
    samples %= (44100 * 4);
    uint32 sectors = samples / 2352;

    char buf[32];
    (void)sprintf(buf, "%02d:%02d:%02d", minutes, seconds, sectors);
    return buf;
}


static void make_toc
(
    vector<uint32>& samples,
    const string& tocName = "data.toc"
)
{
    ofstream out(tocName.c_str());

    out << "CD_DA\n\n" << endl;

    cerr << "calctoc -p ";
    uint32 cur_start = 0;
    for (unsigned int i = 0; i < samples.size(); ++i)
    {
	// desired values:
	uint32 cur_run = 4 * samples[i];
	uint32 run = cur_run;
	uint32 start = cur_start;

	// padded values:
	uint32 padcor = start % 2352;
	start -= padcor;        // padded start time
	run += padcor;
	padcor = run % 2352;
	run -= padcor;          // padded run time

	out << "// Track " << i+1 << endl;
	out << "TRACK AUDIO\n" 
	    << "COPY\n" 
	    << "NO PRE_EMPHASIS\n" 
	    << "TWO_CHANNEL_AUDIO\n" 
	    << "ISRC \"000000000000\"" << endl;
	out << "FILE \"data.wav\" " 
	     << samples_to_toc(start) << " " 
	     << samples_to_toc(run) << endl;

	out << "\n\n" << endl;

	cur_start += cur_run;
	double cur_time = (double)cur_start / (double)(44100 * 4);
	cerr << " " << IOConvert::timeToString(cur_time);
    }
    cerr << endl;
} // end of process_files()


//----------------------------------------------------------------
// copy specified parts of all input files into output file
// On Error: throws an exception
//----------------------------------------------------------------
static void process_files
(
    const FileSettings& files, 
    const AudioFormat& fmt,
    const string& outputFileName,
    const string& tocName = ""
)
{
    if (2 < verbose) cerr << "process files" << endl;
    if (3 < verbose) cerr << "create output file \"" << outputFileName
			  << "\":" << endl;
    IORiffWaveFile out(outputFileName.c_str(), fmt.channelNum,
		       fmt.sampleSize, fmt.sampleRate,
		       fmt.sampleNum);

    vector<uint32> partSamples;
    // handle each file:
    for (unsigned int i=0; i < files.getNum(); i++)
    {
	try 
	{
	    double silence = files.getSilenceTime(i);
	    if (silence > 0)
	    {
		write_silence(silence, out, outputFileName); 
		partSamples.push_back(out.getSampleNum(silence));
	    }
	    else
	    {
		read_one_file(files, i, out, outputFileName, 
			      partSamples); 
	    }
	}
	catch(ErrorClass& error)
	{
	    cerr << error.getMsg() << endl;
	}
    }
    out.writeHeader();
    if (("" != tocName) && partSamples.size())
	make_toc(partSamples);
} // end of process_files()


// --- global functions ---


int main(int argc, char* argv[])
{
    //------------------------------
    // handle command line options:
    //------------------------------
    FileSettings files;
    string tocName;
    int option;
    while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
    {
	switch (option) 
	{
	case 'f': files.addFadeIn(optarg);      break;
	case 'F': files.addFadeOut(optarg);     break;
	case 'i': files.addFileName(optarg);    break;
	case 's': files.addSampleRange(optarg); break;
	case 'S': files.addSilence(optarg);     break;
	case 't': files.addTimeRange(optarg);   break;
	case 'T': tocName = optarg;             break;
	case 'v': verbose++;                    break;
	case 'h': usage(true);                  exit(0);
	case 'V': print_version();              exit(0);
	default : usage();                      exit(1);
	}
    }
    if (0 == files.getNum()) 
    {
	cerr << "Error: missing input file" << endl;
	usage();
	exit(1);
    }
    if (optind >= argc) 
    {
	cerr << "Error: missing output file" << endl;
	usage();
	exit(1);
    }
    // the last argument is the output file name;
    string outputFileName(argv[optind++]);
    // check for too many arguments:
    if (optind < argc) 
    {
	cerr << "Error: too many arguments (last valid is \"" 
	     << outputFileName << "\")" << endl;
	exit (1);
    }

    //--------------------
    // check input files:
    //--------------------
    AudioFormat fmt;
    try { check_input_files(files, fmt); }
    catch(ErrorClass& error)
    {
	cerr << error.getMsg() << endl;
	exit(1);
    }

    //--------------------------
    // process all input files:
    //--------------------------
    try { process_files(files, fmt, outputFileName, tocName); }
    catch(ErrorClass& error)
    {
	cerr << error.getMsg() << endl;
	exit(1);
    }

    exit (0);
    return  (0); // to keep MSVC quiet
} // end of main()


// --- end of file ---
