//---------------------------------------------------------------------------
// Copyright 1997 Andreas Huennebeck
//
// $Source: /home/ah/CvsTree/daten/src/include/AH/Etech/PhysUnit.h,v $
// $Id: PhysUnit.h,v 1.7 2004/04/20 22:04:48 ah Exp $
//
//---------------------------------------------------------------------------

#ifndef ETECH_PHYSUNIT_H	// prevent multiple includes
#define ETECH_PHYSUNIT_H

// --- necessary includes ---

#include <AH/Config/Platform.h>
#include <cmath>
#include <AH/Etech/Types.h>

USING_NAMESPACE(Ah)

// --- types and classes ---

BEGIN_NAMESPACE(Ah)

/// represents a scalar physical unit
/**
 {\tt PhysUnitDbl} is the base class for all scalar physical units, e.g
 temperature, distances and so forth.
 */
class PhysUnitDbl
{
  protected:
    Dbl		Value;
    const char*	Unit;

  public:
    /// constructor
    PhysUnitDbl (const char* u, Dbl v = 0.0)
	: Value(v), Unit(u) {}
    /// constructor
    PhysUnitDbl (const char* u, const Dbl &val)
	: Value(val), Unit(u) {}
    /// copy constructor
    PhysUnitDbl (const PhysUnitDbl& pud) 
	: Value(pud.Value), Unit(pud.Unit) {}

    /// destructor
    virtual ~PhysUnitDbl () {} 

    /// return value of unit
    virtual Dbl getVal () const 
	{ return Value; }

    /// overloaded operator for output into streams
    friend std::ostream& operator << (std::ostream &out, const PhysUnitDbl &pud);

    /// return absolute of value of a {\tt PhysUnitDbl} object
    friend Dbl abs (const PhysUnitDbl &pud) { return std::abs(pud.Value); }
    /// return argument of value  of a {\tt PhysUnitDbl} object (here always 0)
    friend Dbl arg (const PhysUnitDbl &) { return 0.0; }
};


/// represents a complex physical unit
/**
 {\tt PhysUnitCplx} is the base class for all complex physical units, e.g
 voltage, current, impedances and so forth.
 */
class PhysUnitCplx
{
  protected:
    Cplx	Value;
    const char*	Unit;

  public:
    /// constructor
    PhysUnitCplx (const char* u, Dbl r = 0.0, Dbl i = 0.0) 
	: Value(r,i), Unit(u) {}
    /// constructor
    PhysUnitCplx (const char* u, const Cplx &val) 
	: Value(val), Unit(u) {}
    /// copy constructor
    PhysUnitCplx (const PhysUnitCplx& puc) 
	: Value(puc.Value) {}

    /// return {\tt true} if value is infinity
    bool isInf() const 
	{ return Value.isInf(); }
    /// return {\tt true} if value is null
    bool isNul() const 
	{ return Value.isNul(); }
    /// return {\tt true} if value is valid (neither null nor infinity)
    bool isValid() const 
	{ return Value.isValid(); }

    /// return value of unit
    Cplx getVal () const 
	{ return Value; }

    /// overloaded operator for output into streams
    friend std::ostream& operator << (std::ostream &out, const PhysUnitCplx &puc);

    /// overloaded operator for addition of two {\tt PhysUnitCplx} objects
    friend Cplx operator + (const PhysUnitCplx &puc1,
			    const PhysUnitCplx &puc2)
	{ return (puc1.Value + puc2.Value); }
    /// overloaded operator for addition of {\tt PhysUnitCplx} + {\tt Cplx}
    friend Cplx operator + (const PhysUnitCplx &puc,
			    const Cplx &dc)
	{ return (puc.Value + dc); }
    /// overloaded operator for addition of {\tt Cplx} + {\tt PhysUnitCplx}
    friend Cplx operator + (const Cplx &dc,
			    const PhysUnitCplx &puc)
	{ return (dc + puc.Value); }

    /// overloaded operator for subtraction of two {\tt PhysUnitCplx} objects
    friend Cplx operator - (const PhysUnitCplx &puc1,
			    const PhysUnitCplx &puc2)
	{ return (puc1.Value - puc2.Value); }
    /// overloaded operator for subtraction of {\tt PhysUnitCplx} - {\tt Cplx}
    friend Cplx operator - (const PhysUnitCplx &puc,
			    const Cplx &dc)
	{ return (puc.Value - dc); }
    /// overloaded operator for subtraction of {\tt Cplx} - {\tt PhysUnitCplx}
    friend Cplx operator - (const Cplx &dc,
			    const PhysUnitCplx &puc)
	{ return (dc - puc.Value); }

    /// overloaded operator for multiplication of two {\tt PhysUnitCplx} objects
    friend Cplx operator * (const PhysUnitCplx &puc1,
			    const PhysUnitCplx &puc2)
	{ return (puc1.Value * puc2.Value); }
    /// overloaded operator for multiplication of {\tt PhysUnitCplx} * {\tt Cplx}
    friend Cplx operator * (const PhysUnitCplx &puc,
			    const Cplx &dc)
	{ return (puc.Value * dc); }
    /// overloaded operator for multiplication of {\tt Cplx} * {\tt PhysUnitCplx}
    friend Cplx operator * (const Cplx &dc,
			    const PhysUnitCplx &puc)
	{ return (dc * puc.Value); }

    /// overloaded operator for division of two {\tt PhysUnitCplx} objects
    friend Cplx operator / (const PhysUnitCplx &puc1,
			    const PhysUnitCplx &puc2)
	{ return (puc1.Value / puc2.Value); }
    /// overloaded operator for division of {\tt PhysUnitCplx} / {\tt Cplx}
    friend Cplx operator / (const PhysUnitCplx &puc,
			    const Cplx &dc)
	{ return (puc.Value / dc); }
    /// overloaded operator for division of {\tt Cplx} / {\tt PhysUnitCplx}
    friend Cplx operator / (const Cplx &dc,
			    const PhysUnitCplx &puc)
	{ return (dc / puc.Value); }

    /// return absolute of value of a {\tt PhysUnitCplx} object
    friend Dbl abs (const PhysUnitCplx &puc) { return abs(puc.Value); }
    /// return argument of value of a {\tt PhysUnitCplx} object
    friend Dbl arg (const PhysUnitCplx &puc) { return arg(puc.Value); }
};


END_NAMESPACE

#endif // ETECH_PHYSUNIT_H
